@@ -115,9 +115,30 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #ifndef elf_backend_add_symbol_hook
 #define elf_backend_add_symbol_hook	0
 #endif
+#ifndef elf_backend_link_output_symbol_hook
+#define elf_backend_link_output_symbol_hook 0
+#endif
+#ifndef elf_backend_create_dynamic_sections
+#define elf_backend_create_dynamic_sections 0
+#endif
+#ifndef elf_backend_check_relocs
+#define elf_backend_check_relocs	0
+#endif
+#ifndef elf_backend_adjust_dynamic_symbol
+#define elf_backend_adjust_dynamic_symbol 0
+#endif
+#ifndef elf_backend_size_dynamic_sections
+#define elf_backend_size_dynamic_sections 0
+#endif
 #ifndef elf_backend_relocate_section
 #define elf_backend_relocate_section	0
 #endif
+#ifndef elf_backend_finish_dynamic_symbol
+#define elf_backend_finish_dynamic_symbol	0
+#endif
+#ifndef elf_backend_finish_dynamic_sections
+#define elf_backend_finish_dynamic_sections	0
+#endif
 #ifndef elf_backend_begin_write_processing
 #define elf_backend_begin_write_processing	0
 #endif
@@ -151,14 +172,21 @@ static CONST struct elf_backend_data elf32_bed =
   elf_backend_fake_sections,
   elf_backend_section_from_bfd_section,
   elf_backend_add_symbol_hook,
+  elf_backend_link_output_symbol_hook,
+  elf_backend_create_dynamic_sections,
+  elf_backend_check_relocs,
+  elf_backend_adjust_dynamic_symbol,
+  elf_backend_size_dynamic_sections,
   elf_backend_relocate_section,
+  elf_backend_finish_dynamic_symbol,
+  elf_backend_finish_dynamic_sections,
   elf_backend_begin_write_processing,
   elf_backend_final_write_processing,
   elf_backend_ecoff_debug_swap
 };
 
 #ifdef TARGET_BIG_SYM
-bfd_target TARGET_BIG_SYM =
+const bfd_target TARGET_BIG_SYM =
 {
   /* name: identify kind of target */
   TARGET_BIG_NAME,
@@ -174,7 +202,7 @@ bfd_target TARGET_BIG_SYM =
 
   /* object_flags: mask of all file flags */
   (HAS_RELOC | EXEC_P | HAS_LINENO | HAS_DEBUG | HAS_SYMS | HAS_LOCALS |
-   DYNAMIC | WP_TEXT),
+   DYNAMIC | WP_TEXT | D_PAGED),
   
   /* section_flags: mask of all section flags */
   (SEC_HAS_CONTENTS | SEC_ALLOC | SEC_LOAD | SEC_RELOC | SEC_READONLY |
@@ -193,7 +221,7 @@ bfd_target TARGET_BIG_SYM =
      FIXME:  this really has nothing to do with ELF, this is a characteristic
      of the archiver and should be independently tunable.  This value is
      a WAG (wild a** guess) */
-  15,
+  14,
 
   /* align_power_min: minimum alignment restriction for any section
      FIXME:  this value may be target machine dependent */
@@ -246,7 +274,7 @@ bfd_target TARGET_BIG_SYM =
 #endif
 
 #ifdef TARGET_LITTLE_SYM
-bfd_target TARGET_LITTLE_SYM =
+const bfd_target TARGET_LITTLE_SYM =
 {
   /* name: identify kind of target */
   TARGET_LITTLE_NAME,
@@ -262,7 +290,7 @@ bfd_target TARGET_LITTLE_SYM =
 
   /* object_flags: mask of all file flags */
   (HAS_RELOC | EXEC_P | HAS_LINENO | HAS_DEBUG | HAS_SYMS | HAS_LOCALS |
-   DYNAMIC | WP_TEXT),
+   DYNAMIC | WP_TEXT | D_PAGED),
   
   /* section_flags: mask of all section flags */
   (SEC_HAS_CONTENTS | SEC_ALLOC | SEC_LOAD | SEC_RELOC | SEC_READONLY |
@@ -281,7 +309,7 @@ bfd_target TARGET_LITTLE_SYM =
      FIXME:  this really has nothing to do with ELF, this is a characteristic
      of the archiver and should be independently tunable.  This value is
      a WAG (wild a** guess) */
-  15,
+  14,
 
   /* align_power_min: minimum alignment restriction for any section
      FIXME:  this value may be target machine dependent */
