@@ -43,9 +43,10 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "libnlm.h"
 #include "nlmconv.h"
 
-/* Needed for Alpha support.  */
+#ifdef NLMCONV_ALPHA
 #include "coff/sym.h"
 #include "coff/ecoff.h"
+#endif
 
 /* If strerror is just a macro, we want to use the one from libiberty
    since it will handle undefined values.  */
@@ -132,12 +133,6 @@ static void copy_sections PARAMS ((bfd *, asection *, PTR));
 static void mangle_relocs PARAMS ((bfd *, asection *, arelent ***,
 				   long *, char *,
 				   bfd_size_type));
-static void i386_mangle_relocs PARAMS ((bfd *, asection *, arelent ***,
-					long *, char *,
-					bfd_size_type));
-static void alpha_mangle_relocs PARAMS ((bfd *, asection *, arelent ***,
-					 long *, char *,
-					 bfd_size_type));
 static void default_mangle_relocs PARAMS ((bfd *, asection *, arelent ***,
 					   long *, char *,
 					   bfd_size_type));
@@ -146,6 +141,26 @@ static const char *choose_temp_base_try PARAMS ((const char *,
 						 const char *));
 static void choose_temp_base PARAMS ((void));
 static int pexecute PARAMS ((char *, char *[]));
+
+#ifdef NLMCONV_I386
+static void i386_mangle_relocs PARAMS ((bfd *, asection *, arelent ***,
+					long *, char *,
+					bfd_size_type));
+#endif
+
+#ifdef NLMCONV_ALPHA
+static void alpha_mangle_relocs PARAMS ((bfd *, asection *, arelent ***,
+					 long *, char *,
+					 bfd_size_type));
+#endif
+
+#ifdef NLMCONV_POWERPC
+static void powerpc_build_stubs PARAMS ((bfd *, bfd *, asymbol ***, long *));
+static void powerpc_resolve_stubs PARAMS ((bfd *, bfd *));
+static void powerpc_mangle_relocs PARAMS ((bfd *, asection *, arelent ***,
+					   long *, char *,
+					   bfd_size_type));
+#endif
 
 /* The main routine.  */
 
@@ -394,6 +409,16 @@ main (argc, argv)
   if (! bfd_set_section_flags (outbfd, secsec, SEC_HAS_CONTENTS))
     bfd_fatal ("set .nlmsections flags");
 
+#ifdef NLMCONV_POWERPC
+  /* For PowerPC NetWare we need to build stubs for calls to undefined
+     symbols.  Because each stub requires an entry in the TOC section
+     which must be at the same location as other entries in the TOC
+     section, we must do this before determining where the TOC section
+     goes in setup_sections.  */
+  if (bfd_get_arch (inbfd) == bfd_arch_powerpc)
+    powerpc_build_stubs (inbfd, outbfd, &symbols, &symcount);
+#endif
+
   /* Set up the sections.  */
   bfd_map_over_sections (inbfd, setup_sections, (PTR) outbfd);
 
@@ -497,7 +522,7 @@ main (argc, argv)
 	 symbols will confuse it.  */
       if ((sym->flags & BSF_DEBUGGING) == 0
 	  && bfd_asymbol_name (sym)[0] == '_'
-	  && bfd_get_section (sym) == &bfd_und_section)
+	  && bfd_is_und_section (bfd_get_section (sym)))
 	{
 	  if (strcmp (bfd_asymbol_name (sym), "_edata") == 0)
 	    {
@@ -509,6 +534,21 @@ main (argc, argv)
 	      sym->section = bss_sec;
 	      endsym = sym;
 	    }
+
+#ifdef NLMCONV_POWERPC
+	  /* For PowerPC NetWare, we define __GOT0.  This is the start
+	     of the .got section.  */
+	  if (bfd_get_arch (inbfd) == bfd_arch_powerpc
+	      && strcmp (bfd_asymbol_name (sym), "__GOT0") == 0)
+	    {
+	      asection *got_sec;
+
+	      got_sec = bfd_get_section_by_name (inbfd, ".got");
+	      assert (got_sec != (asection *) NULL);
+	      sym->value = got_sec->output_offset;
+	      sym->section = got_sec->output_section;
+	    }
+#endif
  	}
 
       /* If this is a global symbol, check the export list.  */
@@ -560,7 +600,7 @@ main (argc, argv)
 
       /* If it's an undefined symbol, see if it's on the import list.
 	 Change the prefix if necessary.  */
-      if (bfd_get_section (sym) == &bfd_und_section)
+      if (bfd_is_und_section (bfd_get_section (sym)))
 	{
 	  register struct string_list *l;
 
@@ -851,6 +891,12 @@ main (argc, argv)
       strncpy (version_hdr->stamp, "VeRsIoN#", 8);
     }
 
+#ifdef NLMCONV_POWERPC
+  /* Resolve the stubs we build for PowerPC NetWare.  */
+  if (bfd_get_arch (inbfd) == bfd_arch_powerpc)
+    powerpc_resolve_stubs (inbfd, outbfd);
+#endif
+
   /* Copy over the sections.  */
   bfd_map_over_sections (inbfd, copy_sections, (PTR) outbfd);
 
@@ -1093,14 +1139,24 @@ select_output_format (arch, mach, bigendian)
 {
   switch (arch)
     {
+#ifdef NLMCONV_I386
     case bfd_arch_i386:
       return "nlm32-i386";
+#endif
+#ifdef NLMCONV_SPARC
     case bfd_arch_sparc:
       return "nlm32-sparc";
+#endif
+#ifdef NLMCONV_ALPHA
     case bfd_arch_alpha:
       return "nlm32-alpha";
+#endif
+#ifdef NLMCONV_POWERPC
+    case bfd_arch_powerpc:
+      return "nlm32-powerpc";
+#endif
     default:
-      fprintf (stderr, "%s: no default NLM format for %s\n",
+      fprintf (stderr, "%s: support not compiled in for %s\n",
 	       program_name, bfd_printable_arch_mach (arch, mach));
       exit (1);
       /* Avoid warning.  */
@@ -1165,7 +1221,8 @@ setup_sections (inbfd, insec, data_ptr)
 				      bfd_section_alignment (inbfd, insec)))
     bfd_fatal ("set section alignment");
 
-  if (! bfd_set_section_flags (outbfd, outsec, f))
+  if (! bfd_set_section_flags (outbfd, outsec,
+			       f | bfd_get_section_flags (outbfd, outsec)))
     bfd_fatal ("set section flags");
 
   bfd_set_reloc (outbfd, outsec, (arelent **) NULL, 0);
@@ -1306,14 +1363,24 @@ mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr, contents,
 {
   switch (bfd_get_arch (outbfd))
     {
+#ifdef NLMCONV_I386
     case bfd_arch_i386:
       i386_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr,
 			  contents, contents_size);
       break;
+#endif
+#ifdef NLMCONV_ALPHA
     case bfd_arch_alpha:
       alpha_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr,
 			   contents, contents_size);
       break;
+#endif
+#ifdef NLMCONV_POWERPC
+    case bfd_arch_powerpc:
+      powerpc_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr,
+			     contents, contents_size);
+      break;
+#endif
     default:
       default_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr,
 			     contents, contents_size);
@@ -1347,6 +1414,8 @@ default_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr, contents,
 	(*relocs)->address += insec->output_offset;
     }
 }
+
+#ifdef NLMCONV_I386
 
 /* NetWare on the i386 supports a restricted set of relocs, which are
    different from those used on other i386 targets.  This routine
@@ -1468,7 +1537,7 @@ i386_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr, contents,
 
 	  /* Adjust the reloc for the changes we just made.  */
 	  rel->addend = 0;
-	  if (bfd_get_section (sym) != &bfd_und_section)
+	  if (! bfd_is_und_section (bfd_get_section (sym)))
 	    rel->sym_ptr_ptr = bfd_get_section (sym)->symbol_ptr_ptr;
 	}
 
@@ -1503,6 +1572,10 @@ i386_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr, contents,
     }
 }
 
+#endif /* NLMCONV_I386 */
+
+#ifdef NLMCONV_ALPHA
+
 /* On the Alpha the first reloc for every section must be a special
    relocs which hold the GP address.  Also, the first reloc in the
    file must be a special reloc which holds the address of the .lita
@@ -1564,7 +1637,7 @@ alpha_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr, contents,
 	}
 
       *relocs = (arelent *) xmalloc (sizeof (arelent));
-      (*relocs)->sym_ptr_ptr = bfd_abs_section.symbol_ptr_ptr;
+      (*relocs)->sym_ptr_ptr = bfd_abs_section_ptr->symbol_ptr_ptr;
       (*relocs)->address = nlm_alpha_backend_data (outbfd)->lita_address;
       (*relocs)->addend = nlm_alpha_backend_data (outbfd)->lita_size + 1;
       (*relocs)->howto = &nlm32_alpha_nw_howto;
@@ -1578,7 +1651,7 @@ alpha_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr, contents,
       bfd_ecoff_get_gp_value (insec->owner);
 
   *relocs = (arelent *) xmalloc (sizeof (arelent));
-  (*relocs)->sym_ptr_ptr = bfd_abs_section.symbol_ptr_ptr;
+  (*relocs)->sym_ptr_ptr = bfd_abs_section_ptr->symbol_ptr_ptr;
   (*relocs)->address = nlm_alpha_backend_data (outbfd)->gp;
   (*relocs)->addend = 0;
   (*relocs)->howto = &nlm32_alpha_nw_howto;
@@ -1599,6 +1672,425 @@ alpha_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr, contents,
 	(*relocs)->address += insec->output_offset;
     }
 }
+
+#endif /* NLMCONV_ALPHA */
+
+#ifdef NLMCONV_POWERPC
+
+/* We keep a linked list of stubs which we must build.  Because BFD
+   requires us to know the sizes of all sections before we can set the
+   contents of any, we must figure out which stubs we want to build
+   before we can actually build any of them.  */
+
+struct powerpc_stub
+{
+  /* Next stub in linked list.  */
+  struct powerpc_stub *next;
+
+  /* Symbol whose value is the start of the stub.  This is a symbol
+     whose name begins with `.'.  */
+  asymbol *start;
+
+  /* Symbol we are going to create a reloc against.  This is a symbol
+     with the same name as START but without the leading `.'.  */
+  asymbol *reloc;
+
+  /* The TOC index for this stub.  This is the index into the TOC
+     section at which the reloc is created.  */
+  unsigned int toc_index;
+};
+
+/* The linked list of stubs.  */
+
+static struct powerpc_stub *powerpc_stubs;
+
+/* This is what a stub looks like.  The first instruction will get
+   adjusted with the correct TOC index.  */
+
+static unsigned long powerpc_stub_insns[] =
+{
+  0x81820000,		/* lwz	 r12,0(r2) */
+  0x90410014,		/* stw	 r2,20(r1) */
+  0x800c0000,		/* lwz	 r0,0(r12) */
+  0x804c0004,		/* lwz	 r2,r(r12) */
+  0x7c0903a6,		/* mtctr r0 */
+  0x4e800420,		/* bctr */
+  0,			/* Traceback table.  */
+  0xc8000,
+  0
+};
+
+#define POWERPC_STUB_INSN_COUNT \
+  (sizeof powerpc_stub_insns / sizeof powerpc_stub_insns[0])
+
+#define POWERPC_STUB_SIZE (4 * POWERPC_STUB_INSN_COUNT)
+
+/* Each stub uses a four byte TOC entry.  */
+#define POWERPC_STUB_TOC_ENTRY_SIZE (4)
+
+/* The original size of the .got section.  */
+static bfd_size_type powerpc_initial_got_size;
+
+/* Look for all undefined symbols beginning with `.', and prepare to
+   build a stub for each one.  */
+
+static void
+powerpc_build_stubs (inbfd, outbfd, symbols_ptr, symcount_ptr)
+     bfd *inbfd;
+     bfd *outbfd;
+     asymbol ***symbols_ptr;
+     long *symcount_ptr;
+{
+  asection *stub_sec;
+  asection *got_sec;
+  unsigned int got_base;
+  long i;
+  long symcount;
+  long stubcount;
+
+  /* Make a section to hold stubs.  We don't set SEC_HAS_CONTENTS for
+     the section to prevent copy_sections from reading from it.  */
+  stub_sec = bfd_make_section (inbfd, ".stubs");
+  if (stub_sec == (asection *) NULL
+      || ! bfd_set_section_flags (inbfd, stub_sec,
+				  (SEC_CODE
+				   | SEC_RELOC
+				   | SEC_ALLOC
+				   | SEC_LOAD))
+      || ! bfd_set_section_alignment (inbfd, stub_sec, 2))
+    bfd_fatal (".stubs");
+
+  /* Get the TOC section, which is named .got.  */
+  got_sec = bfd_get_section_by_name (inbfd, ".got");
+  if (got_sec == (asection *) NULL)
+    {
+      got_sec = bfd_make_section (inbfd, ".got");
+      if (got_sec == (asection *) NULL
+	  || ! bfd_set_section_flags (inbfd, got_sec,
+				      (SEC_DATA
+				       | SEC_RELOC
+				       | SEC_ALLOC
+				       | SEC_LOAD
+				       | SEC_HAS_CONTENTS))
+	  || ! bfd_set_section_alignment (inbfd, got_sec, 2))
+	bfd_fatal (".got");
+    }
+
+  powerpc_initial_got_size = bfd_section_size (inbfd, got_sec);
+  got_base = powerpc_initial_got_size;
+  got_base = (got_base + 3) &~ 3;
+
+  stubcount = 0;
+
+  symcount = *symcount_ptr;
+  for (i = 0; i < symcount; i++)
+    {
+      asymbol *sym;
+      asymbol *newsym;
+      char *newname;
+      struct powerpc_stub *item;
+
+      sym = (*symbols_ptr)[i];
+
+      /* We must make a stub for every undefined symbol whose name
+	 starts with '.'.  */
+      if (bfd_asymbol_name (sym)[0] != '.'
+	  || ! bfd_is_und_section (bfd_get_section (sym)))
+	continue;
+
+      /* Make a new undefined symbol with the same name but without
+	 the leading `.'.  */
+      newsym = (asymbol *) xmalloc (sizeof (asymbol));
+      *newsym = *sym;
+      newname = (char *) xmalloc (strlen (bfd_asymbol_name (sym)));
+      strcpy (newname, bfd_asymbol_name (sym) + 1);
+      newsym->name = newname;
+
+      /* Define the `.' symbol to be in the stub section.  */
+      sym->section = stub_sec;
+      sym->value = stubcount * POWERPC_STUB_SIZE;
+      /* We set the BSF_DYNAMIC flag here so that we can check it when
+	 we are mangling relocs.  FIXME: This is a hack.  */
+      sym->flags = BSF_LOCAL | BSF_DYNAMIC;
+
+      /* Add this stub to the linked list.  */
+      item = (struct powerpc_stub *) xmalloc (sizeof (struct powerpc_stub));
+      item->start = sym;
+      item->reloc = newsym;
+      item->toc_index = got_base + stubcount * POWERPC_STUB_TOC_ENTRY_SIZE;
+
+      item->next = powerpc_stubs;
+      powerpc_stubs = item;
+      
+      ++stubcount;
+    }
+
+  if (stubcount > 0)
+    {
+      asymbol **s;
+      struct powerpc_stub *l;
+
+      /* Add the new symbols we just created to the symbol table.  */
+      *symbols_ptr = (asymbol **) xrealloc ((char *) *symbols_ptr,
+					    ((symcount + stubcount)
+					     * sizeof (asymbol)));
+      *symcount_ptr += stubcount;
+      s = &(*symbols_ptr)[symcount];
+      for (l = powerpc_stubs; l != (struct powerpc_stub *) NULL; l = l->next)
+	*s++ = l->reloc;
+
+      /* Set the size of the .stubs section and increase the size of
+	 the .got section.  */
+      if (! bfd_set_section_size (inbfd, stub_sec,
+				  stubcount * POWERPC_STUB_SIZE)
+	  || ! bfd_set_section_size (inbfd, got_sec,
+				     (got_base
+				      + (stubcount
+					 * POWERPC_STUB_TOC_ENTRY_SIZE))))
+	bfd_fatal ("stub section sizes");
+    }
+}
+
+/* Resolve all the stubs for PowerPC NetWare.  We fill in the contents
+   of the output section, and create new relocs in the TOC.  */
+
+static void
+powerpc_resolve_stubs (inbfd, outbfd)
+     bfd *inbfd;
+     bfd *outbfd;
+{
+  bfd_byte buf[POWERPC_STUB_SIZE];
+  unsigned int i;
+  unsigned int stubcount;
+  arelent **relocs;
+  asection *got_sec;
+  arelent **r;
+  struct powerpc_stub *l;
+
+  if (powerpc_stubs == (struct powerpc_stub *) NULL)
+    return;
+
+  for (i = 0; i < POWERPC_STUB_INSN_COUNT; i++)
+    bfd_put_32 (outbfd, (bfd_vma) powerpc_stub_insns[i], buf + i * 4);
+
+  got_sec = bfd_get_section_by_name (inbfd, ".got");
+  assert (got_sec != (asection *) NULL);
+  assert (got_sec->output_section->orelocation == (arelent **) NULL);
+
+  stubcount = 0;
+  for (l = powerpc_stubs; l != (struct powerpc_stub *) NULL; l = l->next)
+    ++stubcount;
+  relocs = (arelent **) xmalloc (stubcount * sizeof (arelent *));
+
+  r = relocs;
+  for (l = powerpc_stubs; l != (struct powerpc_stub *) NULL; l = l->next)
+    {
+      arelent *reloc;
+
+      /* Adjust the first instruction to use the right TOC index.  */
+      bfd_put_32 (outbfd, (bfd_vma) powerpc_stub_insns[0] + l->toc_index, buf);
+
+      /* Write this stub out.  */
+      if (! bfd_set_section_contents (outbfd,
+				      bfd_get_section (l->start),
+				      buf,
+				      l->start->value,
+				      POWERPC_STUB_SIZE))
+	bfd_fatal ("writing stub");
+
+      /* Create a new reloc for the TOC entry.  */
+      reloc = (arelent *) xmalloc (sizeof (arelent));
+      reloc->sym_ptr_ptr = &l->reloc;
+      reloc->address = l->toc_index + got_sec->output_offset;
+      reloc->addend = 0;
+      reloc->howto = bfd_reloc_type_lookup (inbfd, BFD_RELOC_32);
+				      
+      *r++ = reloc;
+    }
+
+  bfd_set_reloc (outbfd, got_sec->output_section, relocs, stubcount);
+}
+
+/* Adjust relocation entries for PowerPC NetWare.  We do not output
+   TOC relocations.  The object code already contains the offset from
+   the TOC pointer.  When the function is called, the TOC register,
+   r2, will be set to the correct TOC value, so there is no need for
+   any further reloc.  */
+
+/*ARGSUSED*/
+static void
+powerpc_mangle_relocs (outbfd, insec, relocs_ptr, reloc_count_ptr, contents,
+		       contents_size)
+     bfd *outbfd;
+     asection *insec;
+     register arelent ***relocs_ptr;
+     long *reloc_count_ptr;
+     char *contents;
+     bfd_size_type contents_size;
+{
+  const reloc_howto_type *toc_howto;
+  long reloc_count;
+  register arelent **relocs;
+  register long i;
+
+  toc_howto = bfd_reloc_type_lookup (insec->owner, BFD_RELOC_PPC_TOC16);
+  if (toc_howto == (reloc_howto_type *) NULL)
+    abort ();
+
+  /* If this is the .got section, clear out all the contents beyond
+     the initial size.  We must do this here because copy_sections is
+     going to write out whatever we return in the contents field.  */
+  if (strcmp (bfd_get_section_name (insec->owner, insec), ".got") == 0)
+    memset (contents + powerpc_initial_got_size, 0,
+	    (bfd_get_section_size_after_reloc (insec)
+	     - powerpc_initial_got_size));
+
+  reloc_count = *reloc_count_ptr;
+  relocs = *relocs_ptr;
+  for (i = 0; i < reloc_count; i++)
+    {
+      arelent *rel;
+      asymbol *sym;
+      bfd_vma sym_value;
+
+      rel = *relocs++;
+      sym = *rel->sym_ptr_ptr;
+
+      /* Convert any relocs against the .bss section into relocs
+         against the .data section.  */
+      if (strcmp (bfd_get_section_name (outbfd, bfd_get_section (sym)),
+		  NLM_UNINITIALIZED_DATA_NAME) == 0)
+	{
+	  asection *datasec;
+
+	  datasec = bfd_get_section_by_name (outbfd,
+					     NLM_INITIALIZED_DATA_NAME);
+	  if (datasec != NULL)
+	    {
+	      rel->addend += (bfd_get_section_vma (outbfd,
+						   bfd_get_section (sym))
+			      + sym->value);
+	      rel->sym_ptr_ptr = datasec->symbol_ptr_ptr;
+	      sym = *rel->sym_ptr_ptr;
+	    }
+	}
+
+      /* We must be able to resolve all PC relative relocs at this
+	 point.  If we get a branch to an undefined symbol we build a
+	 stub, since NetWare will resolve undefined symbols into a
+	 pointer to a function descriptor.  */
+      if (rel->howto->pc_relative)
+	{
+	  /* This check for whether a symbol is in the same section as
+	     the reloc will be wrong if there is a PC relative reloc
+	     between two sections both of which were placed in the
+	     same output section.  This should not happen.  */
+	  if (bfd_get_section (sym) != insec->output_section)
+	    fprintf (stderr, "%s: unresolved PC relative reloc against %s\n",
+		     program_name, bfd_asymbol_name (sym));
+	  else
+	    {
+	      bfd_vma val;
+
+	      assert (rel->howto->size == 2 && rel->howto->pcrel_offset);
+	      val = bfd_get_32 (outbfd, (bfd_byte *) contents + rel->address);
+	      val = ((val &~ rel->howto->dst_mask)
+		     | (((val & rel->howto->src_mask)
+			 + (sym->value - rel->address)
+			 + rel->addend)
+			& rel->howto->dst_mask));
+	      bfd_put_32 (outbfd, val, (bfd_byte *) contents + rel->address);
+
+	      /* If this reloc is against an stubbed symbol and the
+		 next instruction is
+		     cror 31,31,31
+		 then we replace the next instruction with
+		     lwz  r2,20(r1)
+		 This reloads the TOC pointer after a stub call.  */
+	      if (bfd_asymbol_name (sym)[0] == '.'
+		  && (sym->flags & BSF_DYNAMIC) != 0
+		  && (bfd_get_32 (outbfd,
+				  (bfd_byte *) contents + rel->address + 4)
+		      == 0x4ffffb82)) /* cror 31,31,31 */
+		bfd_put_32 (outbfd, (bfd_vma) 0x80410014, /* lwz r2,20(r1) */
+			    (bfd_byte *) contents + rel->address + 4);
+
+	      --*reloc_count_ptr;
+	      --relocs;
+	      memmove (relocs, relocs + 1,
+		       (size_t) ((reloc_count - 1) * sizeof (arelent *)));
+	      continue;
+	    }
+	}
+
+      /* When considering a TOC reloc, we do not want to include the
+	 symbol value.  The symbol will be start of the TOC section
+	 (which is named .got).  We do want to include the addend.  */
+      if (rel->howto == toc_howto)
+	sym_value = 0;
+      else
+	sym_value = sym->value;
+
+      /* If this is a relocation against a symbol with a value, or
+	 there is a reloc addend, we need to update the addend in the
+	 object file.  */
+      if (sym_value + rel->addend != 0)
+	{
+	  bfd_vma val;
+
+	  switch (rel->howto->size)
+	    {
+	    case 1:
+	      val = bfd_get_16 (outbfd,
+				(bfd_byte *) contents + rel->address);
+	      val = ((val &~ rel->howto->dst_mask)
+		     | (((val & rel->howto->src_mask)
+			 + sym_value
+			 + rel->addend)
+			& rel->howto->dst_mask));
+	      if ((bfd_signed_vma) val < - 0x8000
+		  || (bfd_signed_vma) val >= 0x8000)
+		fprintf (stderr,
+			 "%s: overflow when adjusting relocation against %s\n",
+			 program_name, bfd_asymbol_name (sym));
+	      bfd_put_16 (outbfd, val, (bfd_byte *) contents + rel->address);
+	      break;
+
+	    case 2:
+	      val = bfd_get_32 (outbfd,
+				(bfd_byte *) contents + rel->address);
+	      val = ((val &~ rel->howto->dst_mask)
+		     | (((val & rel->howto->src_mask)
+			 + sym_value
+			 + rel->addend)
+			& rel->howto->dst_mask));
+	      bfd_put_32 (outbfd, val, (bfd_byte *) contents + rel->address);
+	      break;
+
+	    default:
+	      abort ();
+	    }
+
+	  rel->sym_ptr_ptr = bfd_get_section (sym)->symbol_ptr_ptr;
+	  rel->addend = 0;
+	}
+
+      /* Now that we have incorporated the addend, remove any TOC
+	 relocs.  */
+      if (rel->howto == toc_howto)
+	{
+	  --*reloc_count_ptr;
+	  --relocs;
+	  memmove (relocs, relocs + 1,
+		   (size_t) ((reloc_count - i) * sizeof (arelent *)));
+	  continue;
+	}
+
+      rel->address += insec->output_offset;
+    }
+}
+
+#endif /* NLMCONV_POWERPC */
 
 /* Name of linker.  */
 #ifndef LD_NAME
@@ -1628,7 +2120,7 @@ link_inputs (inputs, ld)
   for (q = inputs; q != NULL; q = q->next)
     ++c;
 
-  argv = (char **) alloca (c + 5);
+  argv = (char **) alloca ((c + 5) * sizeof(char *));
 
 #ifndef __MSDOS__
   if (ld == NULL)
@@ -1660,7 +2152,7 @@ link_inputs (inputs, ld)
   sprintf (unlink_on_exit, "%s.O", temp_filename);
 
   argv[0] = ld;
-  argv[1] = (char *) "-r";
+  argv[1] = (char *) "-Ur";
   argv[2] = (char *) "-o";
   argv[3] = unlink_on_exit;
   i = 4;
