@@ -37,11 +37,14 @@
 
 #include "config.h"
 
-/* This is the code recommended in the autoconf documentation --
+/* This is the code recommended in the autoconf documentation, almost
    verbatim.  If it doesn't work for you, let me know, and notify
    djm@gnu.ai.mit.edu as well.  */
+/* Added #undef for DJ Delorie.  KR 1994.10.21 */
+
 /* AIX requires this to be the first thing in the file.  */
 #ifdef __GNUC__
+# undef alloca
 # define alloca __builtin_alloca
 #else
 # if HAVE_ALLOCA_H
