@@ -40,21 +40,28 @@ if [ "x${LIB_PATH}" = "x" ] ; then
       fi
    else
       # Cross.
-      LIB_PATH=:
+      LIB_PATH=
    fi
 fi
+
+# Always search $(tooldir)/lib, aka /usr/local/TARGET/lib.
+LIB_PATH=${LIB_PATH}:`echo ${libdir} | sed -e s'|/lib$||'`/${target_alias}/lib
+
 LIB_SEARCH_DIRS=`echo ${LIB_PATH} | tr ':' ' ' | sed -e 's/\([^ ][^ ]*\)/SEARCH_DIR(\1);/g'`
 
-# Generate 5 script files from a master script template in
-# ${srcdir}/scripttempl/${SCRIPT_NAME}.sh.  Which one of the 5 script files
-# is actually used depends on command line options given to ld.
-# (SCRIPT_NAME was set in the emulparams_file.)
+# Generate 5 or 6 script files from a master script template in
+# ${srcdir}/scripttempl/${SCRIPT_NAME}.sh.  Which one of the 5 or 6
+# script files is actually used depends on command line options given
+# to ld.  (SCRIPT_NAME was set in the emulparams_file.)
 #
 # A .x script file is the default script.
 # A .xr script is for linking without relocation (-r flag).
 # A .xu script is like .xr, but *do* create constructors (-Ur flag).
 # A .xn script is for linking with -n flag (mix text and data on same page).
 # A .xbn script is for linking with -N flag (mix text and data on same page).
+# A .xs script is for generating a shared library with the --shared
+#   flag; it is only generated if $GENERATE_SHLIB_SCRIPT is set by the
+#   emulation parameters.
 
 SEGMENT_SIZE=${SEGMENT_SIZE-${PAGE_SIZE}}
 
@@ -96,6 +103,15 @@ DATA_ALIGNMENT=${DATA_ALIGNMENT_N}
 (. ${srcdir}/scripttempl/${SCRIPT_NAME}.sc) | sed -e '/^ *$/d' > \
   ldscripts/${EMULATION_NAME}.xbn
 
+if test -n "$GENERATE_SHLIB_SCRIPT"; then
+  LD_FLAG=shared
+  DATA_ALIGNMENT=${DATA_ALIGNMENT_s-${DATA_ALIGNMENT_}}
+  CREATE_SHLIB=" "
+  # Note that TEXT_START_ADDR is set to NONPAGED_TEXT_START_ADDR.
+  (. ${srcdir}/scripttempl/${SCRIPT_NAME}.sc) | sed -e '/^ *$/d' > \
+    ldscripts/${EMULATION_NAME}.xs
+fi
+
 test "$DEFAULT_EMULATION" = "$EMULATION_NAME" && COMPILE_IN=true
 
 # Generate em_${EMULATION_NAME}.c.
